from django.apps import apps
from django.db import models
from django.forms import modelform_factory
from django.http.request import HttpRequest
from django.shortcuts import redirect, render

def make_admin_list_view(model: models.Model):
    def view(request: HttpRequest):
        if not request.user.is_superuser:
            raise PermissionError("Superuser access required.")

        # pagination
        # search
        # ordering
        # ... CRUD subclass.
        
        return render(request, "onto/crud/list.html", {
            "model_meta": model._meta,
            "objects": model.objects.all()
        })

    return view

def make_admin_edit_view(model: models.Model):
    def view(request: HttpRequest, pk=None):
        if not request.user.is_superuser:
            raise PermissionError("Superuser access required.")
        
        if not hasattr(view, "Form"):
            view.Form = modelform_factory(model, exclude=[])

        instance = model.objects.filter(pk=pk).first()
        form = view.Form(request.POST or None, instance=instance)

        if request.method == "POST" and form.is_valid():
            instance = form.save()
            return redirect(f'onto_crud:edit-{model._meta.app_label}-{model._meta.model_name}', instance.pk)
        
        return render(request, "onto/crud/create.html", {
            "model_meta": model._meta,
            "object": instance,
            "form": form
        })

    return view


def htmx(view):
    """Adds attributes to the request object based on the HTMX context."""
    def middleware(request):
        request.htmx = ("HX-Request" in request.headers)
        request.modal = ("HX-Modal" in request.headers)
        return view(request)
    return middleware


def base_template(request):
    template = "dash/base.html"
    if request.htmx:
        if request.modal:
            template = "dash/base_modal.html"
        else:
            template = "dash/base_ajax.html"

    print(f"base_template: {template}")
    return {"base_template": template}


from collections import defaultdict, namedtuple

from django.db import models
from django.urls import path

from .views import make_admin_edit_view, make_admin_list_view

URLs = namedtuple("URLs", ("pattern_list", "app_namespace"))

def admin_crud(model: models.Model):
    """
    Generates a CRUD GUI for this Model, accessible only to superusers.
    """

    app_label = model._meta.app_label
    model_name = model._meta.model_name

    if model in admin_crud.registry[app_label]:
        raise ValueError(f"{model} has already been registered as a CRUD model.")

    admin_crud.registry[app_label].add(model)

    list_view = make_admin_list_view(model)
    edit_view = make_admin_edit_view(model)
    admin_crud.urls.pattern_list.append(path(f"{app_label}/{model_name}/", list_view, name=f"list-{app_label}-{model_name}"))
    admin_crud.urls.pattern_list.append(path(f"{app_label}/{model_name}/edit", edit_view, name=f"create-{app_label}-{model_name}"))
    admin_crud.urls.pattern_list.append(path(f"{app_label}/{model_name}/edit/<pk>", edit_view, name=f"edit-{app_label}-{model_name}"))

    return model

admin_crud.registry = defaultdict(set)
admin_crud.urls = URLs(
    app_namespace="onto_crud",
    pattern_list=list(), 
)